# This file includes tools to modify the code and make the implementation easier
def expand_sources(source_path):
    # Get all files contained in a source path (recursive)
    from os import walk
    from os.path import join
    files = [join(dp, f) for dp, dn, fn in walk(source_path) for f in fn]
    file_paths = {}
    for f in files:
        name = f.split("/")[-1]
        file_paths[name] = f
    return file_paths


def match_module_and_filename(module, filenames):
    # Gets the filepath of the corresponding module
    import re
    pattern = "%s\.[fFh90]" % module
    matches = [fn for fn in filenames if re.match(pattern, fn)]
    assert len(matches) == 1, "Multiple file match for module %s" % module
    return matches[0]


def find_declaration_line(variable, filepath, vault):
    # Finds line at which the variable is declared
    import re
    from UtilsRPE.Functions import in_which_block
    import warnings
    # Declaration pattern
    # TODO: This might have issues with split lines
    pattern = "::.*\W%s\W" % variable.name
    with open(filepath, "r") as f:
        lines = [l for l in f]

        current_block = vault[variable.routine.module.name].main()

        for _index, _line in enumerate(lines):
            if re.search(pattern, _line):
                if current_block == variable.routine:
                    return _index
            else:
                current_block = in_which_block(_line, current_block, vault=vault)
    warnings.warn("Declaration of variable %s not found in module %s" % (variable.name, variable.routine.module.name))
    return None


def replace_declaration(variable, file_paths, vault, target_precision="dp"):
    # This function will find the file where a variable is declared and will replace the declaration
    # with a new one using the new precision spec.
    files = file_paths.keys()
    module = variable.routine.module.name
    filepath = file_paths[match_module_and_filename(module, files)]
    declaration_line_index = find_declaration_line(variable, filepath, vault)
    if declaration_line_index is not None:
        with open(filepath, "r") as f:
            data = [l for l in f]
            declaration_line = data[declaration_line_index]
            new_line = replace_real_declaration(variable, declaration_line, target_precision=target_precision)
            data[declaration_line_index] = new_line
            updated_text = "".join(data)
        with open(filepath, "w") as output:
            output.write(updated_text)


def replace_real_declaration(variable, declaration_line, target_precision="dp"):
    # This function takes a declaration line (string) and updates it with the new type
    # If the declaration contains more variables it keeps the other variables with the original attributes.
    from AutoRPE.UtilsRPE.Functions import remove_comments, split_declaration
    import re
    import warnings
    declaration_line_without_comments = remove_comments(declaration_line).strip()

    attributes, arguments = split_declaration(declaration_line_without_comments)

    # Real statement pattern
    pattern = "real *\(([^\)]*)\)"

    if len(arguments) == 1:
        if declaration_line_without_comments[-1] == "&":
            new_declaration_line = "! WARNING!! the declaration hasn't been properly fixed!\n" + \
                                   declaration_line
        else:
            m = re.search(pattern, declaration_line_without_comments, re.I)
            if m:
                real_statement = m.group(0)
                precision_specification = m.group(1)
                new_real_statement = real_statement.replace(precision_specification, target_precision)
                new_declaration_line_without_comments = \
                    declaration_line_without_comments.replace(real_statement, new_real_statement)
                new_declaration_line = \
                    declaration_line.replace(declaration_line_without_comments, new_declaration_line_without_comments)
            else:
                warnings.warn("Problems finding real statement:\n%s" % declaration_line)
                new_declaration_line = "! WARNING!! the declaration hasn't been properly fixed!\n" + \
                                       declaration_line
        return new_declaration_line
    else:
        if variable.name not in arguments:
            raise AssertionError("Variable %s not found in list of arguments" % variable.name, attributes)
        m = re.search(pattern, attributes, re.I)
        if m:
            real_statement = m.group(0)
            precision_specification = m.group(1)
            new_real_statement = real_statement.replace(precision_specification, target_precision)
            new_attributes = attributes.replace(real_statement, new_real_statement)
            additional_declaration = "%s :: %s" % (new_attributes, variable.name)
            m = re.search(r"(.*\b)%s(\b.*)" % variable.name, declaration_line_without_comments)
            if m:

                declaration_line_without_selected_var = "".join(m.groups(0))
            else:
                warnings.warn("WTF?")

            # Comma patterns to remove the extra comma
            beginning_pat = "::( *\,)"
            beginning_pat_substitution = "::"
            middle_pat = ",( *,)"
            middle_pat_sub = ","
            last_pat = "(, *)$"
            last_pat_sub = ""

            for pat, sub in [(beginning_pat, beginning_pat_substitution), (middle_pat, middle_pat_sub),
                             (last_pat, last_pat_sub)]:
                m = re.search(pat, declaration_line_without_selected_var)
                if m:
                    new_declaration_line_without_comments = re.sub(pat, sub, declaration_line_without_selected_var)
                    new_declaration_line = declaration_line.replace(declaration_line_without_comments,
                                                                    new_declaration_line_without_comments)
                    break
        else:
            pass

        # Adding proper indentation
        leading_spaces = (len(new_declaration_line) - len(new_declaration_line.lstrip()))
        new_declaration_line = "%s%s\n%s" % (
        leading_spaces * " ", additional_declaration.lstrip(), new_declaration_line)

        return new_declaration_line


def change_code(list_filename, source_path, vault_path):
    # The purpose of this function is to modify the declarations of several variables in a source-code.
    # As inputs it takes the path to the file with the list of variables, the path to the sources that will be modified
    # and the path to the vault database containing information about the code.

    from UtilsRPE.Functions import load_vault, load_variable_list

    vault = load_vault(vault_path)
    variable_ids = load_variable_list(list_filename)
    file_paths = expand_sources(source_path=source_path)

    variables = [vault.get_variable_by_id(_id) for _id in variable_ids]

    for variable in variables:
        replace_declaration(variable, file_paths, vault)

    pass



